import geopandas as gpd

from ra2ce.analysis.analysis_config_data.enums.damage_curve_enum import DamageCurveEnum
from ra2ce.analysis.damages.shape_to_integrate_object.hz_to_integrate_shaper import (
    HzToIntegrateShaper,
)
from ra2ce.analysis.damages.shape_to_integrate_object.man_to_integrate_shaper import (
    ManToIntegrateShaper,
)
from ra2ce.analysis.damages.shape_to_integrate_object.osd_to_integrate_shaper import (
    OsdToIntegrateShaper,
)
from ra2ce.analysis.damages.shape_to_integrate_object.to_Integrate_shaper_protocol import (
    ToIntegrateShaperProtocol,
)


class ToIntegrateShaperFactory:
    @staticmethod
    def get_shaper(
        gdf: gpd.GeoDataFrame, damage_function: DamageCurveEnum
    ) -> ToIntegrateShaperProtocol:
        if damage_function == DamageCurveEnum.HZ:
            return HzToIntegrateShaper(gdf)
        if damage_function == DamageCurveEnum.OSD:
            return OsdToIntegrateShaper(gdf)
        if damage_function == DamageCurveEnum.MAN:
            return ManToIntegrateShaper(gdf)

        raise NotImplementedError(
            f"damage function {damage_function} not yet supported."
        )



"""
                    GNU GENERAL PUBLIC LICENSE
                      Version 3, 29 June 2007

    Risk Assessment and Adaptation for Critical Infrastructure (RA2CE).
    Copyright (C) 2023 Stichting Deltares

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import logging

import geopandas as gpd
import networkx as nx
import numpy as np
import pandas as pd
import pyproj
from rasterstats import point_query, zonal_stats
from tqdm import tqdm

from ra2ce.network import networks_utils as ntu
from ra2ce.network.exporters.network_exporter_factory import NetworkExporterFactory
from ra2ce.network.graph_files.graph_files_collection import GraphFilesCollection
from ra2ce.network.hazard.hazard_common_functions import (
    get_edges_geoms,
    validate_extent_graph,
)
from ra2ce.network.hazard.hazard_files import HazardFiles
from ra2ce.network.hazard.hazard_intersect.hazard_intersect_builder_for_gpkg import (
    HazardIntersectBuilderForGpkg,
)
from ra2ce.network.hazard.hazard_intersect.hazard_intersect_builder_for_table import (
    HazardIntersectBuilderForTable,
)
from ra2ce.network.hazard.hazard_intersect.hazard_intersect_builder_for_tif import (
    HazardIntersectBuilderForTif,
)
from ra2ce.network.network_config_data.network_config_data import NetworkConfigData


class HazardOverlay:
    """Class where the hazard overlay happens.

    Attributes:
        network: GeoDataFrame of the network.
        graph_files: NetworkX graphs.
    """

    _ra2ce_name_key = "RA2CE name"

    def __init__(
        self,
        config: NetworkConfigData,
        graph_files: GraphFilesCollection,
    ):
        # Sections properties
        self._network_file_id = config.network.file_id
        self._output_graph_dir = config.static_path.joinpath("output_graph")
        self._output_dir = config.output_path
        self._origins = config.origins_destinations.origins
        self._destinations = config.origins_destinations.destinations
        self._save_gpkg = config.network.save_gpkg
        self._isolation_locations = config.static_path.joinpath(
            "network", config.isolation.locations
        )
        if config.static_path.joinpath("network", config.isolation.locations).is_file():
            self._isolation_locations = config.static_path.joinpath(
                "network", config.isolation.locations
            )
        else:
            self._isolation_locations = None
        # Hazard properties
        self._hazard_field_name = config.hazard.hazard_field_name
        self._hazard_id = config.hazard.hazard_id
        self._hazard_map = config.hazard.hazard_map
        self._hazard_crs = config.hazard.hazard_crs
        self._hazard_aggregate_wl = config.hazard.aggregate_wl.config_value
        self._hazard_directory = config.static_path.joinpath("hazard")
        self._overlay_segmented_network = config.hazard.overlay_segmented_network

        # graph files
        self.graph_files = graph_files

        # files
        self.hazard_files = HazardFiles.from_hazard_map(self._hazard_map)

        # bookkeeping for the hazard map names
        self.hazard_name_table = self.get_hazard_name_table()
        self.hazard_names = list(
            dict.fromkeys(list(self.hazard_name_table["File name"]))
        )
        self.ra2ce_names = list(
            dict.fromkeys(
                [n[:-3] for n in self.hazard_name_table[self._ra2ce_name_key]]
            )
        )
        logging.info("Initialized hazard object.")

    def od_hazard_intersect(
        self, graph: nx.Graph, ods: gpd.GeoDataFrame
    ) -> tuple[nx.Graph, gpd.GeoDataFrame]:
        """Overlays the origin and destination locations and edges with the hazard maps

        Args:
            graph (NetworkX graph): The origin-destination graph that should be overlayed with the hazard raster(s)

        Returns:
            graph (NetworkX graph): The origin-destination graph hazard raster(s) data joined to both the origin- and
            destination nodes and the edges.
        """
        ## Intersect the origin and destination nodes with the hazard map (now only geotiff possible)
        # Verify the graph type (networkx)
        assert isinstance(graph, nx.Graph)
        extent_graph = ntu.get_graph_edges_extent(graph)

        # Get all node geometries
        od_nodes = [(n, ndata) for n, ndata in graph.nodes.data() if "od_id" in ndata]
        od_ids = [n[0] for n in od_nodes]

        # Get all edge geometries
        edges_geoms = get_edges_geoms(graph)

        for i, (hn, rn) in enumerate(zip(self.hazard_names, self.ra2ce_names)):
            # Check if the hazard and graph extents overlap
            validate_extent_graph(extent_graph, self.hazard_files.tif[i])

            # Read the hazard values at the nodes and write to the nodes.
            tqdm.pandas(desc="Destinations hazard overlay with " + hn)
            _tif_hazard_files = str(self.hazard_files.tif[i])
            flood_stats = ods.geometry.progress_apply(
                lambda x, _file_values=_tif_hazard_files: point_query(x, _file_values)
            )

            flood_stats = flood_stats.apply(lambda x: x[0] if x[0] else 0)

            # Update the ODs GeoDataFrame
            ods[rn + "_" + self._hazard_aggregate_wl[:2]] = flood_stats

            # Update the graph
            attribute_dict = {
                od: {rn + "_" + self._hazard_aggregate_wl[:2]: wl}
                for od, wl in zip(od_ids, flood_stats)
            }
            nx.set_node_attributes(graph, attribute_dict)

            # Read the hazard values at the edges and write to the edges.
            # Add a no-data value for the edges that do not have a geometry
            nx.set_edge_attributes(
                graph,
                {
                    (u, v, k): {rn + "_" + self._hazard_aggregate_wl[:2]: np.nan}
                    for u, v, k, edata in graph.edges.data(keys=True)
                    if "geometry" not in edata
                },
            )

            # Add the hazard values to the edges that do have a geometry
            gdf = gpd.GeoDataFrame(
                {"geometry": [edata["geometry"] for u, v, k, edata in edges_geoms]}
            )
            tqdm.pandas(desc="OD graph hazard overlay with " + hn)
            flood_stats = gdf.geometry.progress_apply(
                lambda x, _files_values=_tif_hazard_files: zonal_stats(
                    x,
                    _files_values,
                    all_touched=True,
                    stats=f"{self._hazard_aggregate_wl}",  # TODO: ADD MEAN WITHOUT THE NANs
                )
            )

            try:
                flood_stats = flood_stats.fillna(0)
                nx.set_edge_attributes(
                    graph,
                    {
                        (edges[0], edges[1], edges[2]): {
                            rn
                            + "_"
                            + self._hazard_aggregate_wl[:2]: x[0][
                                self._hazard_aggregate_wl
                            ]
                        }
                        for x, edges in zip(flood_stats, edges_geoms)
                    },
                )
            except Exception:
                logging.warning(
                    "No aggregation method ('aggregate_wl') is chosen - choose from 'max', 'min' or 'mean'."
                )

            # Get the fraction of the road that is intersecting with the hazard
            tqdm.pandas(desc="OD graph fraction with hazard overlay with " + hn)
            graph_fraction_flooded = gdf.geometry.progress_apply(
                lambda x, _files_values=_tif_hazard_files: ntu.fraction_flooded(
                    x, _files_values
                )
            )
            graph_fraction_flooded = graph_fraction_flooded.fillna(0)
            nx.set_edge_attributes(
                graph,
                {
                    (edges[0], edges[1], edges[2]): {rn + "_fr": x}
                    for x, edges in zip(graph_fraction_flooded, edges_geoms)
                },
            )

        return graph, ods

    def get_point_hazard_from_network(
        self, points: gpd.GeoDataFrame, network: gpd.GeoDataFrame
    ) -> gpd.GeoDataFrame:
        """get hazard informations for points at nearest network locations.

        Args:
            points (gpd.GeoDataFrame): points of interest, will be snapped to the network by nearest distance method.
            network (gpd.GeoDataFrame): network of interest

        Returns:
            gpd.GeoDataFrame: points with relation to network and hazard informations

        Raise:
            error: if node id columns do not exisit in the network
        """

        # TODO: use the same convention for shapefiles and osm
        if all(
            c_idx in network.columns for c_idx in ["node_A", "node_B"]
        ):  # shapefiles
            network["edge_fid"] = [
                f"{na}_{nb}" for na, nb in network[["node_A", "node_B"]].values
            ]
        elif all(c_idx in network.columns for c_idx in ["u", "v"]):  # osm
            network["edge_fid"] = [
                f"{na}_{nb}" for na, nb in network[["u", "v"]].values
            ]
        else:
            logging.error(
                "Node id columns are not found in the network. Support node_A, node_B or u, v"
            )

        # check for crs
        if points.crs.is_geographic:
            crs = points.crs
            points = points.to_crs(3857)
            network = network.to_crs(points.crs)

        points = gpd.sjoin_nearest(
            points, network, how="left", distance_col="edges_distance"
        )
        if any(points["edges_distance"].isna()) > 0:
            logging.warning("Not all points are snapped to the network.")

        return points.to_crs(crs)

    def get_hazard_name_table(self) -> pd.DataFrame:
        all_agg_types = {
            "max": "maximum",
            "min": "minimum",
            "mean": "average",
            "fr": "fraction of network segment impacted by hazard",
        }
        chosen_agg_types = [
            all_agg_types[self._hazard_aggregate_wl],
            all_agg_types["fr"],
        ]
        df = pd.DataFrame()
        df[["File name", "Aggregation method"]] = [
            (haz.stem, agg_type)
            for haz in self._hazard_map
            for agg_type in chosen_agg_types
        ]
        if all(["RP" in haz.stem for haz in self._hazard_map]):
            # Return period hazard maps are used
            # Note: no hazard type is indicated because the name became too long
            rps = [haz.stem.split("RP_")[-1].split("_")[0] for haz in self._hazard_map]
            df[self._ra2ce_name_key] = [
                "RP" + rp + "_" + agg_type[:2]
                for rp in rps
                for agg_type in chosen_agg_types
            ]
        else:
            # Event hazard maps are used
            # Note: no hazard type is indicated because the name became too long
            df[self._ra2ce_name_key] = [
                "EV" + str(i + 1) + "_" + agg_type[:2]
                for i in range(len(self._hazard_map))
                for agg_type in chosen_agg_types
            ]
        df["Full path"] = [haz for haz in self._hazard_map for _ in chosen_agg_types]
        return df

    def hazard_intersect(
        self, to_overlay: gpd.GeoDataFrame | nx.Graph
    ) -> gpd.GeoDataFrame | nx.Graph:
        """Handler function that chooses the right function for overlaying the network with the hazard data."""
        # To improve performance we need to initialize the variables
        if self.hazard_files.tif:
            return HazardIntersectBuilderForTif(
                hazard_aggregate_wl=self._hazard_aggregate_wl,
                hazard_names=self.hazard_names,
                ra2ce_names=self.ra2ce_names,
                hazard_tif_files=self.hazard_files.tif,
            ).get_intersection(to_overlay)
        elif self.hazard_files.gpkg:
            return HazardIntersectBuilderForGpkg(
                hazard_field_name=self._hazard_field_name,
                hazard_aggregate_wl=self._hazard_aggregate_wl,
                ra2ce_names=self.ra2ce_names,
                hazard_gpkg_files=self.hazard_files.gpkg,
            ).get_intersection(to_overlay)
        elif self.hazard_files["table"]:
            return HazardIntersectBuilderForTable(
                hazard_field_name=self._hazard_field_name,
                network_file_id=self._network_file_id,
                hazard_id=self._hazard_id,
                ra2ce_name_key=self._ra2ce_name_key,
            ).get_intersection(to_overlay)

        raise ValueError(
            f"The overlay of the combination of hazard file(s) '{self.hazard_files}' and network type '{type(to_overlay)}' is not available."
            f"Please check your input data."
        )

    def get_reproject_graph(
        self,
        original_graph: nx.Graph,
        in_crs: pyproj.CRS,
        out_crs: pyproj.CRS,
    ) -> nx.Graph:
        """Reproject networkX graph"""
        extent_graph = ntu.get_graph_edges_extent(original_graph)
        logging.info("Graph extent before reprojecting: %s", extent_graph)

        graph_reprojected = ntu.reproject_graph(original_graph, in_crs, out_crs)
        extent_graph_reprojected = ntu.get_graph_edges_extent(graph_reprojected)
        logging.info("Graph extent after reprojecting: %s", extent_graph_reprojected)

        return graph_reprojected

    def get_original_geoms_graph(self, graph_original: nx.MultiGraph, graph_new):
        original_geometries = nx.get_edge_attributes(graph_original, "geometry")
        _graph_new = graph_new.copy()
        nx.set_edge_attributes(_graph_new, original_geometries, "geometry")
        return _graph_new.copy()

    def _export_network_files(self, graph_type: str, types_to_export: list[str]):
        _exporter = NetworkExporterFactory()
        _exporter.export(
            network=self.graph_files.get_graph(graph_type),
            basename=graph_type,
            output_dir=self._output_graph_dir,
            export_types=types_to_export,
        )
        self.graph_files.set_file(_exporter.get_pickle_path())

    def load_origins_destinations(self):
        od_path = self._output_graph_dir.joinpath("origin_destination_table.feather")
        return gpd.read_feather(od_path)

    def _create_base_overlay(self, base_graph: nx.MultiGraph) -> nx.MultiGraph:

        # Check if the graph needs to be reprojected
        _hazard_crs = pyproj.CRS.from_user_input(self._hazard_crs)
        _graph_crs = pyproj.CRS.from_user_input(
            "EPSG:4326"
        )  # this is WGS84, TODO: Make flexible by including in the network ini

        if _hazard_crs != _graph_crs:
            # Temporarily reproject the graph to the CRS of the hazard
            logging.warning(
                """Hazard crs %s and graph crs %s are inconsistent, we try to reproject the graph crs""",
                _hazard_crs,
                _graph_crs,
            )
            _graph_reprojected = self.get_reproject_graph(
                base_graph, _graph_crs, _hazard_crs
            )

            # Do the actual hazard intersect
            _base_graph_hazard_reprojected = self.hazard_intersect(_graph_reprojected)

            # Assign the original geometries to the reprojected raster
            return self.get_original_geoms_graph(
                base_graph, _base_graph_hazard_reprojected
            )

        return self.hazard_intersect(base_graph)

    def _create_origin_destinations_overlay(
        self, origin_destinations_graph: nx.MultiGraph
    ) -> tuple[nx.MultiGraph, gpd.GeoDataFrame]:
        _ods = self.load_origins_destinations()

        # Check if the graph needs to be reprojected
        _hazard_crs = pyproj.CRS.from_user_input(self._hazard_crs)
        _graph_crs = pyproj.CRS.from_user_input(
            "EPSG:4326"
        )  # this is WGS84, TODO: Make flexible by including in the network ini

        if (
            _hazard_crs != _graph_crs
        ):  # Temporarily reproject the graph to the CRS of the hazard
            logging.warning(
                """Hazard crs %s and graph crs %s are inconsistent, we try to reproject the graph crs""",
                _hazard_crs,
                _graph_crs,
            )
            if _hazard_crs != _ods.crs:
                logging.warning(
                    """Hazard crs %s and OD crs %s are inconsistent, we try to reproject the graph crs""",
                    _hazard_crs,
                    _ods.crs,
                )
                _ods_reprojected = _ods.to_crs(_hazard_crs)

            _graph_reprojected = self.get_reproject_graph(
                origin_destinations_graph, _graph_crs, _hazard_crs
            )

            # Do the actual hazard intersect
            (
                _od_graph_hazard_reprojected,
                _ods_hazard_reprojected,
            ) = self.od_hazard_intersect(_graph_reprojected, _ods_reprojected)

            # Assign the original geometries to the reprojected dataset
            _graph_hazard = self.get_original_geoms_graph(
                origin_destinations_graph, _od_graph_hazard_reprojected
            )
            _ods = _ods_hazard_reprojected.to_crs(_ods.crs)
        else:
            (
                _graph_hazard,
                _ods,
            ) = self.od_hazard_intersect(origin_destinations_graph, _ods)

        return (_graph_hazard, _ods)

    def _create_base_network_overlay(
        self, base_network: gpd.GeoDataFrame
    ) -> gpd.GeoDataFrame:
        logging.info("Iterating overlay of GeoPandas Dataframe.")

        # Check if the graph needs to be reprojected
        _hazard_crs = pyproj.CRS.from_user_input(self._hazard_crs)
        _gdf_crs = pyproj.CRS.from_user_input(base_network.crs)

        if (
            _hazard_crs != _gdf_crs
        ):  # Temporarily reproject the graph to the CRS of the hazard
            logging.warning(
                "Hazard crs %s and gdf crs %s are inconsistent, we try to reproject the gdf crs",
                _hazard_crs,
                _gdf_crs,
            )

            _extent_gdf = base_network.total_bounds
            logging.info("Gdf extent before reprojecting: %s", _extent_gdf)

            _gdf_reprojected = base_network.copy().to_crs(_hazard_crs)
            _extent_gdf_reprojected = _gdf_reprojected.total_bounds
            logging.info("Gdf extent after reprojecting: %s", _extent_gdf_reprojected)

            # Do the actual hazard intersect
            _gdf_reprojected = self.hazard_intersect(_gdf_reprojected)

            # Assign the original geometries to the reprojected raster
            _original_geometries = base_network["geometry"]
            _gdf_reprojected["geometry"] = _original_geometries
            return _gdf_reprojected.copy()

        # read previously created file
        logging.info("Setting 'base_network_hazard' graph.")
        return self.hazard_intersect(base_network)

    def _create_isolated_locations_overlay(
        self, base_network_hazard: gpd.GeoDataFrame
    ) -> gpd.GeoDataFrame:
        logging.info("Detected isolated locations, checking for hazard overlay.")
        _locations = gpd.read_file(self._isolation_locations, engine="pyogrio")
        _locations["i_id"] = _locations.index

        # get hazard at locations from network based on nearest
        logging.info("Get hazard at locations from network.")
        _locations_hazard = self.get_point_hazard_from_network(
            _locations, base_network_hazard
        )
        return _locations_hazard

    def create(self):
        """Overlays the different possible graph and network objects with the hazard data

        Arguments:
            None
            (implicit) : (self.graphs_files) : GraphFile object with the graph names and files as attributes

        Returns:
            self.graph_files : same object, but with new graphs _hazard, which are copies of the graphs but with hazard data

        Effect:
            write all the objects

        """
        types_to_export = ["pickle"] if not self._save_gpkg else ["pickle", "gpkg"]

        if (
            not self.graph_files.base_graph.file
            and not self.graph_files.origins_destinations_graph.file
        ):
            logging.warning(
                "Either a base graph or OD graph is missing to intersect the hazard with. "
                "Check your network folder."
            )

        #### Step 1: hazard overlay of the base graph (NetworkX) ###
        if (
            self.graph_files.base_graph.file
            and not self.graph_files.base_graph_hazard.file
        ):
            self.graph_files.base_graph_hazard.graph = self._create_base_overlay(
                self.graph_files.base_graph.get_graph()
            )

            self._export_network_files("base_graph_hazard", types_to_export)

        #### Step 2: hazard overlay of the origins_destinations (NetworkX) ###
        if (
            self.graph_files.origins_destinations_graph.file
            and self._origins
            and self._destinations
            and not self.graph_files.origins_destinations_graph_hazard.file
        ):
            (
                self.graph_files.origins_destinations_graph_hazard.graph,
                _ods,
            ) = self._create_origin_destinations_overlay(
                self.graph_files.origins_destinations_graph.get_graph()
            )

            # Save graphs/network with hazard
            self._export_network_files(
                "origins_destinations_graph_hazard", types_to_export
            )

            # Save the OD pairs (GeoDataFrame) as pickle
            _od_table_feather = self._output_graph_dir.joinpath(
                "origin_destination_table.feather"
            )
            _ods.to_feather(
                _od_table_feather,
                index=False,
            )
            logging.info(
                "Saved %s in %s.", _od_table_feather.name, _od_table_feather.parent
            )

            # Save the OD pairs (GeoDataFrame) as shapefile
            if self._save_gpkg:
                _ods_path = self._output_graph_dir.joinpath(
                    "origin_destination_table.gpkg"
                )
                _ods.to_file(_ods_path, index=False)
                logging.info(
                    "Saved %s in %s.", _ods_path.stem, _ods_path.resolve().parent
                )

        #### Step 3: iterate overlay of the GeoPandas Dataframe (if any) ###
        if (
            self.graph_files.base_network.file
            and not self.graph_files.base_network_hazard.file
            and (self._overlay_segmented_network or self._isolation_locations)
        ):
            self.graph_files.base_network_hazard.graph = (
                self._create_base_network_overlay(
                    self.graph_files.base_network.get_graph()
                )
            )

            # Save segmented network with hazard
            self._export_network_files("base_network_hazard", types_to_export)

        #### Step 4: hazard overlay of the locations that are checked for isolation ###
        if self._isolation_locations:
            self.graph_files.locations_hazard.graph = (
                self._create_isolated_locations_overlay(
                    self.graph_files.base_network_hazard.get_graph()
                )
            )

            # Save isolated locations with hazard
            self._export_network_files("locations_hazard", "pickle")

        # Save the hazard name bookkeeping table.
        self.hazard_name_table.to_excel(
            self._output_dir.joinpath("hazard_names.xlsx"), index=False
        )

        return self.graph_files



"""
                    GNU GENERAL PUBLIC LICENSE
                      Version 3, 29 June 2007

    Risk Assessment and Adaptation for Critical Infrastructure (RA2CE).
    Copyright (C) 2023 Stichting Deltares

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path

import pandas as pd

from ra2ce.analysis.analysis_config_wrapper import AnalysisConfigWrapper

FILE_NAME_KEY = "File name"
RA2CE_NAME_KEY = "RA2CE name"


@dataclass
class HazardNames:
    names_df: pd.DataFrame

    @classmethod
    def from_file(cls, hazard_names_file: Path) -> HazardNames:
        """
        Create a HazardNames object from a file.

        Args:
            hazard_names_file (Path): Path to the files with the hazard names.

        Returns:
            HazardNames: HazardNames object.
        """
        if hazard_names_file and hazard_names_file.is_file():
            _names_df = pd.read_excel(hazard_names_file)
        else:
            _names_df = pd.DataFrame(data=None)
        return cls(names_df=_names_df)

    @classmethod
    def from_config(cls, analysis_config: AnalysisConfigWrapper) -> HazardNames:
        """
        Create a HazardNames object from an analysis configuration.

        Args:
            analysis_config (AnalysisConfigWrapper): Analysis configuration.

        Returns:
            HazardNames: HazardNames object.
        """
        if analysis_config.config_data.output_path:
            _hazard_file = analysis_config.config_data.output_path.joinpath(
                "hazard_names.xlsx"
            )
        else:
            _hazard_file = None
        return cls.from_file(_hazard_file)

    @property
    def names(self) -> list[str]:
        """
        Get the list of hazard names.

        Returns:
            list[str]: List of hazard names.
        """
        if self.names_df.empty:
            return []
        return self.names_df[FILE_NAME_KEY].tolist()

    def get_name(self, hazard: str) -> str:
        """
        Get the RA2CE name of a specific hazard.

        Args:
            hazard (str): Name of the hazard.

        Returns:
            str: RA2CE name of the hazard.
        """
        return self.names_df.loc[
            self.names_df[FILE_NAME_KEY] == hazard,
            RA2CE_NAME_KEY,
        ].values[0]



from ra2ce.analysis.analysis_config_data.enums.weighing_enum import WeighingEnum
from ra2ce.analysis.losses.weighing_analysis.length_weighing_analysis import (
    LengthWeighingAnalysis,
)
from ra2ce.analysis.losses.weighing_analysis.time_weighing_analysis import (
    TimeWeighingAnalysis,
)
from ra2ce.analysis.losses.weighing_analysis.weighing_analysis_protocol import (
    WeighingAnalysisProtocol,
)


class WeighingAnalysisFactory:
    @staticmethod
    def get_analysis(weighing_type: WeighingEnum) -> WeighingAnalysisProtocol:
        if weighing_type == WeighingEnum.TIME:
            return TimeWeighingAnalysis()
        if weighing_type == WeighingEnum.LENGTH:
            return LengthWeighingAnalysis()

        raise NotImplementedError(
            "Weighing type {} not yet supported.".format(weighing_type)
        )



"""
                    GNU GENERAL PUBLIC LICENSE
                      Version 3, 29 June 2007

    Risk Assessment and Adaptation for Critical Infrastructure (RA2CE).
    Copyright (C) 2023 Stichting Deltares

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import logging
from pathlib import Path

import pandas as pd

from ra2ce.analysis.damages.damage.damage_fraction_uniform import DamageFractionUniform
from ra2ce.analysis.damages.damage.max_damage import MaxDamageByRoadTypeByLane


class DamageFunctionByRoadTypeByLane:
    """
    A damage function that has different max damages per road type, but a uniform damage_fraction curve


    The attributes need to be of the type:
    self.max_damage (MaxDamage_byRoadType_byLane)
    self.damage_fraction (DamageFractionHazardSeverityUniform)

    """

    def __init__(
        self,
        max_damage: MaxDamageByRoadTypeByLane = None,
        damage_fraction: DamageFractionUniform = None,
        name: str = "",
        hazard: str = "flood",
        type: str = "depth_damage",
        infra_type: str = "road",
    ):
        # Construct using the parent class __init__
        self.name = name
        self.hazard = hazard
        self.type = type
        self.infra_type = infra_type
        self.max_damage = max_damage  # Should be a MaxDamage object
        self.damage_fraction = (
            damage_fraction  # Should be a DamageFractionHazardSeverity object
        )
        self.prefix = None  # Should be two characters long at maximum

    def set_prefix(self):
        self.prefix = self.name[0:2]
        logging.info(
            "The prefix: '{}' refers to curve name '{}' in the results".format(
                self.prefix, self.name
            )
        )

    def from_input_folder(self, folder_path: Path):
        """Construct a set of damage functions from csv files located in the folder_path

        Arguments:
            *folder_path* (Pathlib Path) : path to folder where csv files can be found
        """

        def find_unique_csv_file(folder_path: Path, part_of_filename: str) -> Path:
            """
            Find unique csv file in a folder, with a given part_of_filename
            Raises a warning if no file can be found, and an error if more than one file is found
            """
            result = []
            for file in folder_path.iterdir():
                if (part_of_filename in file.stem) and (file.suffix == ".csv"):
                    result.append(file)
            if len(result) > 1:
                raise ValueError(
                    "Found more then one damage file in {}".format(folder_path)
                )
            elif len(result) == 0:
                raise ValueError(
                    "Did not find any damage file in {}".format(folder_path)
                )

            return result[0]

        # Load the max_damage object
        max_damage = MaxDamageByRoadTypeByLane()
        max_dam_path = find_unique_csv_file(folder_path, "max_damage")
        max_damage.from_csv(max_dam_path, sep=";")

        self.max_damage = max_damage

        # Load the damage fraction function
        # search in the folder for something *damage_fraction
        damage_fraction = DamageFractionUniform()
        dam_fraction_path = find_unique_csv_file(folder_path, "hazard_severity")
        damage_fraction.from_csv(dam_fraction_path, sep=";")
        self.damage_fraction = damage_fraction

        damage_fraction.create_interpolator()

    # Todo: these two below functions are maybe better implemented at a lower level?
    def add_max_damage(self, df: pd.DataFrame, prefix: str = None):
        """ "Ads the max damage value to the dataframe"""
        cols = df.columns
        assert "road_type" in cols, "no column 'road type' in df"
        assert "lanes" in cols, "no column 'lanes in df"

        max_damage_data = self.max_damage.data
        df["{}_temp_max_dam".format(prefix)] = max_damage_data.lookup(
            df["infra_type"], df["lanes"]
        )
        return df

    def calculate_damage(
        self,
        df: pd.DataFrame,
        damage_function_prefix: str,
        hazard_prefix: str,
        event_prefix: str,
    ) -> pd.DataFrame:
        """
        Calculates the damage for one event. The prefixes are used to find/set the right df columns.

        Args:
            df (pd.DataFrame): dataframe with road network data.
            damage_function_prefix (str): prefix to identify the right damage function e.g. 'A'.
            hazard_prefix (str): prefix to identify the right hazard e.g. 'F'.
            event_prefix (str): prefix to identify the right event, e.g. 'EV1'

        Returns:
            pd.DataFrame: dataframe data with the damage calculation added as new column
        """

        interpolator = (
            self.damage_fraction.interpolator
        )  # get the interpolator function

        # Find correct columns in dataframe
        result_col = "dam_{}_{}".format(event_prefix, damage_function_prefix)
        max_dam_col = "{}_temp_max_dam".format(damage_function_prefix)
        hazard_severity_col = "{}_{}_me".format(
            hazard_prefix, event_prefix
        )  # mean is hardcoded now
        hazard_fraction_col = "{}_{}_fr".format(
            hazard_prefix, event_prefix
        )  # fraction column is hardcoded

        df[result_col] = round(
            df[max_dam_col].astype(float)  # max damage (euro/m)
            * interpolator(df[hazard_severity_col].astype(float))  # damage curve  (-)
            * df["length"]  # segment length (m)
            * df[hazard_fraction_col],
            0,
        )  # round to whole numbers
        return df



"""
                    GNU GENERAL PUBLIC LICENSE
                      Version 3, 29 June 2007

    Risk Assessment and Adaptation for Critical Infrastructure (RA2CE).
    Copyright (C) 2023 Stichting Deltares

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import logging
import re
import time

import geopandas as gpd
import networkx as nx
import numpy as np
import pandas as pd
from networkx import MultiDiGraph

from ra2ce.analysis.analysis_collection import AnalysisCollection
from ra2ce.analysis.analysis_config_data.enums.event_type_enum import EventTypeEnum
from ra2ce.analysis.analysis_config_wrapper import AnalysisConfigWrapper
from ra2ce.analysis.analysis_result_wrapper import AnalysisResultWrapper
from ra2ce.analysis.analysis_result_wrapper_exporter import (
    AnalysisResultWrapperExporter,
)
from ra2ce.analysis.damages.analysis_damages_protocol import AnalysisDamagesProtocol
from ra2ce.configuration.config_wrapper import ConfigWrapper
from ra2ce.runners.analysis_runner_protocol import AnalysisRunner


class DamagesAnalysisRunner(AnalysisRunner):
    def __str__(self) -> str:
        return "Damages Analysis Runner"

    @staticmethod
    def _update_link_based_values(
        damages_link_based_graph: nx.MultiDiGraph | nx.MultiGraph,
        result_segment_based: pd.DataFrame,
        segment_id_column: str,
        result_column: str,
        segment_values_list: str,
    ) -> nx.MultiDiGraph | nx.MultiGraph:
        """
        Derive the values in the provided graph (link_based) based on the results of the segment-based graph.

        Parameters:
        - damages_link_based_graph (nx.MultiDiGraph | nx.MultiGraph): The graph with link-based damage data.
        - result_segment_based (pd.DataFrame): DataFrame containing segment-based damage data.
        - segment_id_column (str): The column name in both the graph data and DataFrame representing segment IDs.
        - result_column (str): The column name in the graph data and DataFrame representing the damage result.
        - segment_values_list (str): The column name in the graph data to store the segment damage values.

        Returns:
        - damages_link_based_graph (nx.MultiDiGraph | nx.MultiGraph)
        """
        for u, v, key, data in damages_link_based_graph.edges(keys=True, data=True):
            damages_link_based_graph[u][v][key][segment_values_list] = []
            damages_link_based_graph[u][v][key][result_column] = 0

        # Lookup segment_id_list for each edge
        for u, v, key, data in damages_link_based_graph.edges(keys=True, data=True):
            link_value = data[result_column]
            segment_id_list = (
                data[segment_id_column]
                if isinstance(data[segment_id_column], list)
                else [data[segment_id_column]]
            )

            # Read value for each segment_id & append to segment_values_list and calculate link_value
            for segment_id in segment_id_list:
                segment_value = result_segment_based.loc[
                    result_segment_based[segment_id_column] == segment_id,
                    result_column,
                ].squeeze()

                data[segment_values_list].append(round(segment_value, 2))

                if np.isnan(segment_value):
                    segment_value = 0

                link_value += round(segment_value, 2)

            data[result_column] = round(link_value, 2)
        return damages_link_based_graph

    @staticmethod
    def can_run(ra2ce_input: ConfigWrapper) -> bool:
        if (
            not ra2ce_input.analysis_config
            or not ra2ce_input.analysis_config.config_data.damages_list
        ):
            return False
        if not ra2ce_input.network_config:
            return False
        _network_config = ra2ce_input.network_config.config_data
        if not _network_config.hazard or not _network_config.hazard.hazard_map:
            logging.error(
                "Please define a hazard map in your network.ini file. Unable to calculate damages."
            )
            return False
        return True

    @staticmethod
    def _get_result_link_based(
        analysis: AnalysisDamagesProtocol,
        analysis_config: AnalysisConfigWrapper,
        result_segment_based: gpd.GeoDataFrame,
    ) -> gpd.GeoDataFrame:
        # Step 00: define parameters
        base_graph_hazard_graph = analysis_config.graph_files.base_graph_hazard.graph
        damage_curve = analysis.analysis.damage_curve
        segment_id_column = "rfid_c"

        # Find the hazard columns; these may be events or return periods
        event_cols = [
            col
            for col in result_segment_based.columns
            if (col[0].isupper() and col[1] == "_")
        ]
        events = set([x.split("_")[1] for x in event_cols])  # set of unique events

        # Step 0: create a deep copy of the base_graph_hazard to compose further as the final outcome of this process
        damages_link_based_graph = base_graph_hazard_graph.copy()

        # Step 1: Create a new attribute damage_segments_list for each edge
        for event in events:
            if damage_curve == analysis.analysis.damage_curve.HZ:
                damage_result_columns = [
                    f"dam_{event}_{damage_curve}"
                ]  # there is one damage column
            elif damage_curve == analysis.analysis.damage_curve.OSD:
                pattern = rf"dam_.*_{event}_representative"
                damage_result_columns = [
                    col
                    for col in result_segment_based.columns
                    if re.match(pattern, col)
                ]
                # there are multiple damage columns
            elif damage_curve == analysis.analysis.damage_curve.MAN:
                pattern = rf"dam_{event}_.*"
                damage_result_columns = [
                    col
                    for col in result_segment_based.columns
                    if re.match(pattern, col)
                ]  # there are multiple damage columns
            else:
                raise ValueError(f"damage curve {damage_curve} is invalid")

            for damage_result_column in damage_result_columns:
                # Step 2: Get damage of each link.
                # Read damage for each segment_id & append to damage_segments_list and calculate link_damage
                damages_link_based_graph = (
                    DamagesAnalysisRunner._update_link_based_values(
                        damages_link_based_graph=damages_link_based_graph,
                        result_segment_based=result_segment_based,
                        segment_id_column=segment_id_column,
                        result_column=damage_result_column,
                        segment_values_list=f"{damage_result_column}_segments",
                    )
                )

        # Step 3: get risk of each link.
        # Read risk for each segment_id & append to risk_segments_list and calculate link_risk
        if analysis.analysis.event_type == EventTypeEnum.RETURN_PERIOD:
            pattern = r"(risk.*)"
            risk_result_columns = [
                col for col in result_segment_based.columns if re.match(pattern, col)
            ]
            for risk_result_column in risk_result_columns:
                damages_link_based_graph = (
                    DamagesAnalysisRunner._update_link_based_values(
                        damages_link_based_graph=damages_link_based_graph,
                        result_segment_based=result_segment_based,
                        segment_id_column=segment_id_column,
                        result_column=risk_result_column,
                        segment_values_list=f"{risk_result_column}_segments",
                    )
                )

        # Step 4: Convert the edge attributes to a GeoDataFrame
        edge_attributes = []
        for u, v, key, data in damages_link_based_graph.edges(keys=True, data=True):
            edge_attributes.append({**{"u": u, "v": v, "key": key}, **data})
        damages_link_based_gdf = gpd.GeoDataFrame(edge_attributes)
        return damages_link_based_gdf

    def run(
        self, analysis_config: AnalysisConfigWrapper
    ) -> list[AnalysisResultWrapper]:
        _analysis_collection = AnalysisCollection.from_config(analysis_config)
        _results = []
        for analysis in _analysis_collection.damages_analyses:
            logging.info(
                "----------------------------- Started analyzing '%s'  -----------------------------",
                analysis.analysis.name,
            )
            starttime = time.time()

            _result_segmented = analysis.execute()
            _result_link_based = self._get_result_link_based(
                analysis=analysis,
                analysis_config=analysis_config,
                result_segment_based=_result_segmented,
            )
            analysis_name = analysis.analysis.name
            for _result, suffix in zip(
                [_result_segmented, _result_link_based], ["segmented", "link_based"]
            ):
                _result_wrapper = AnalysisResultWrapper(
                    analysis_result=_result, analysis=analysis
                )
                _result_wrapper.analysis.analysis.name = analysis_name + "_" + suffix
                _results.append(_result_wrapper)

                AnalysisResultWrapperExporter().export_result(_result_wrapper)

            endtime = time.time()
            logging.info(
                "----------------------------- Analysis '%s' finished. "
                "Time: %ss  -----------------------------",
                analysis.analysis.name,
                str(round(endtime - starttime, 2)),
            )
        return _results



"""
                    GNU GENERAL PUBLIC LICENSE
                      Version 3, 29 June 2007
    Risk Assessment and Adaptation for Critical Infrastructure (RA2CE).
    Copyright (C) 2023 Stichting Deltares
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
from dataclasses import dataclass, field

from ra2ce.network.network_config_data.enums.road_type_enum import RoadTypeEnum


@dataclass(kw_only=True)
class ResilienceCurves:
    """
    Class to store the resilience curves for different link types.
    """

    resilience_curves: dict[
        tuple[RoadTypeEnum, tuple[float, float]], list[tuple[float, float]]
    ] = field(default_factory=dict)

    @property
    def ranges(self) -> list[tuple[float, float]]:
        return list(set(_key[1] for _key in self.resilience_curves.keys()))

    def has_resilience_curve(
        self, link_type: RoadTypeEnum, hazard_range: tuple[float, float]
    ) -> bool:
        """
        Check if a resilience curve exists for a given link type and hazard range.

        Args:
            link_type (RoadTypeEnum): The type of the link.
            hazard_range (tuple[float, float]): The range of the hazard.

        Returns:
            bool: True if the resilience curve exists.
        """
        return self.get_resilience_curve(link_type, hazard_range) is not None

    def get_resilience_curve(
        self, link_type: RoadTypeEnum, hazard_range: tuple[float, float]
    ) -> list[tuple[float, float]] | None:
        """
        Get the resilience curve for a given link type and hazard range.

        Args:
            link_type (RoadTypeEnum): The type of the link.
            hazard_range (tuple[float, float]): The range of the hazard.

        Returns:
            list[tuple[float, float]]: The resilience curve.
        """
        if (link_type, hazard_range) in self.resilience_curves.keys():
            return self.resilience_curves[(link_type, hazard_range)]
        return None

    def get_duration_steps(
        self, link_type: RoadTypeEnum, hazard_range: tuple[float, float]
    ) -> list[float]:
        """
        Get the duration steps for a given link type and hazard range.

        Args:
            link_type (RoadTypeEnum): The type of the link.
            hazard_range (tuple[float, float]): The range of the hazard.

        Returns:
            list[float]: The duration steps.
        """
        _curves = self.resilience_curves[(link_type, hazard_range)]
        return [curve[0] for curve in _curves]

    def get_functionality_loss_ratio(
        self, link_type: RoadTypeEnum, hazard_range: tuple[float, float]
    ) -> list[float]:
        """
        Get the functionality loss ratio for a given link type and hazard range.

        Args:
            link_type (RoadTypeEnum): The type of the link.
            hazard_range (tuple[float, float]): The range of the hazard.

        Returns:
            list[float]: The functionality loss ratio.
        """
        _curves = self.resilience_curves[(link_type, hazard_range)]
        return [curve[1] for curve in _curves]

    def calculate_disruption(
        self, link_type: RoadTypeEnum, hazard_range: tuple[float, float]
    ) -> float:
        """
        Calculate the disruption for a given link type and hazard range.
        It is guaranteed that the duration steps and functionality loss ratio have the same dimensions.

        Args:
            link_type (RoadTypeEnum): The type of the link.
            hazard_range (tuple[float, float]): The range of the hazard.

        Returns:
            float: The calculated disruption.
        """
        return sum(
            map(
                lambda x, y: x * y,
                self.get_duration_steps(link_type, hazard_range),
                self.get_functionality_loss_ratio(link_type, hazard_range),
            )
        )



"""
                    GNU GENERAL PUBLIC LICENSE
                      Version 3, 29 June 2007

    Risk Assessment and Adaptation for Critical Infrastructure (RA2CE).
    Copyright (C) 2023 Stichting Deltares

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""


from pathlib import Path
from typing import Any, Protocol, runtime_checkable


@runtime_checkable
class Ra2ceExporterProtocol(Protocol):
    def export(self, export_path: Path, export_data: Any) -> None:  # pragma: no cover
        """
        Exports the given data to the given path.

        Args:
            export_path (Path): File path where to save the `export_data`.
            export_data (Any): Data to be exported.
        """
        pass



"""
                    GNU GENERAL PUBLIC LICENSE
                      Version 3, 29 June 2007

    Risk Assessment and Adaptation for Critical Infrastructure (RA2CE).
    Copyright (C) 2023 Stichting Deltares

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
import logging
from dataclasses import dataclass

import networkx as nx
from tqdm import tqdm

from ra2ce.network.network_simplification.network_simplification_with_attribute_exclusion import (
    NetworkSimplificationWithAttributeExclusion,
)
from ra2ce.network.network_simplification.network_simplification_without_attribute_exclusion import (
    NetworkSimplificationWithoutAttributeExclusion,
)

NxGraph = nx.Graph | nx.MultiGraph | nx.MultiDiGraph


@dataclass(kw_only=True)
class NetworkGraphSimplificator:
    """
    Factory dataclass to simplify the containing graph.
    """

    graph_complex: NxGraph
    attributes_to_exclude: list[str]
    new_id: str = "rfid"

    def simplify(
        self,
    ) -> tuple[nx.Graph, nx.Graph, tuple[dict, dict]]:
        """
        Create a simplified graph with unique ids from a complex graph

        Returns:
            tuple[nx.Graph, nx.Graph, tuple[dict, dict]]: The simple and complex graph and the "id" tables.
        """
        logging.info("Simplifying graph")
        try:
            _graph_complex = self._graph_create_unique_ids(
                self.graph_complex, "{}_c".format(self.new_id)
            )
            _graph_simple = self._get_graph_simple()

            # Create look_up_tables between graphs with unique ids
            (
                _simple_to_complex,
                _complex_to_simple,
            ) = self._graph_link_simple_id_to_complex(_graph_simple)

            # Store id table and add simple ids to complex graph
            _id_tables = (_simple_to_complex, _complex_to_simple)
            _graph_complex = self._add_simple_id_to_graph_complex(
                _graph_complex, _complex_to_simple, self.new_id
            )
            logging.info("Simplified graph successfully created")
        except Exception as exc:
            _graph_simple = None
            _id_tables = None
            logging.error("Did not create a simplified version of the graph (%s)", exc)
        return _graph_simple, _graph_complex, _id_tables

    def _get_graph_simple(self) -> NxGraph:
        if any(self.attributes_to_exclude):
            _graph_simple = NetworkSimplificationWithAttributeExclusion(
                nx_graph=self.graph_complex,
                attributes_to_exclude=self.attributes_to_exclude,
            ).simplify_graph()
        else:
            self.graph_complex = (
                self.graph_complex.to_directed()
            )  # simplification function requires nx.MultiDiGraph

            # Create simplified graph and add unique ids
            _graph_simple = NetworkSimplificationWithoutAttributeExclusion(
                nx_graph=self.graph_complex
            ).simplify_graph()

        return self._graph_create_unique_ids(_graph_simple, self.new_id)

    def _graph_create_unique_ids(
        self, graph: nx.Graph, new_id_name: str = "rfid"
    ) -> nx.Graph:
        # Check if new_id_name exists and if unique
        u, v, k = list(graph.edges)[0]
        if new_id_name in graph.edges[u, v, k]:
            return graph
        # TODO: decide if we always add a new ID (in iGraph this is different)
        # if len(set([str(e[-1][new_id_name]) for e in graph.edges.data(keys=True)])) < len(graph.edges()):
        for i, (u, v, k) in enumerate(graph.edges(keys=True)):
            graph[u][v][k][new_id_name] = i + 1
        logging.info("Added a new unique identifier field '%s'.", new_id_name)
        return graph

    def _add_simple_id_to_graph_complex(
        self, complex_graph: nx.classes.Graph, complex_to_simple, new_id
    ) -> nx.classes.Graph:
        """Adds the appropriate ID of the simple graph to each edge of the complex graph as a new attribute 'rfid'

        Arguments:
            complex_graph (Graph) : The complex graph, still lacking 'rfid'
            complex_to_simple (dict) : lookup table linking complex to simple graphs

        Returns:
            complex_graph (Graph) : Same object, with added attribute 'rfid'

        """

        obtained_complex_ids = nx.get_edge_attributes(
            complex_graph, "{}_c".format(new_id)
        )  # {(u,v,k) : 'rfid_c'}
        simple_ids_per_complex_id = obtained_complex_ids  # start with a copy

        for key, value in obtained_complex_ids.items():  # {(u,v,k) : 'rfid_c'}
            try:
                new_value = complex_to_simple[
                    value
                ]  # find simple id belonging to the complex id
                simple_ids_per_complex_id[key] = new_value
            except KeyError as e:
                logging.error(
                    "Could not find the simple ID belonging to complex ID %s; value set to None. Full error: %s",
                    key,
                    e,
                )
                simple_ids_per_complex_id[key] = None

        # Now the format of simple_ids_per_complex_id is: {(u,v,k) : 'rfid}
        nx.set_edge_attributes(complex_graph, simple_ids_per_complex_id, new_id)

        return complex_graph

    def _graph_link_simple_id_to_complex(self, graph_simple: nx.classes.graph.Graph):
        """
        Create lookup tables (dicts) to match edges_ids of the complex and simple graph
        Optionally, saves these lookup tables as json files.

        Arguments:
            graph_simple (Graph) : Graph, containing attribute 'new_id'

        Returns:
            simple_to_complex (dict): Keys are ids of the simple graph, values are lists with all matching complex ids
            complex_to_simple (dict): Keys are the ids of the complex graph, value is the matching simple_ID

        We need this because the simple graph is derived from the complex graph, and therefore initially only the
        simple graph knows from which complex edges it was created. To assign this information also to the complex
        graph we invert the look-up dictionary
        @author: Kees van Ginkel en Margreet van Marle
        """
        # Iterate over the simple, because this already has the corresponding complex information
        lookup_dict = {}
        # keys are the ids of the simple graph, values are lists with all matching complex id's
        for u, v, k in tqdm(graph_simple.edges(keys=True)):
            key_1 = graph_simple[u][v][k]["{}".format(self.new_id)]
            value_1 = graph_simple[u][v][k]["{}_c".format(self.new_id)]
            lookup_dict[key_1] = value_1

        inverted_lookup_dict = {}
        # keys are the ids of the complex graph, value is the matching simple_ID
        for key, value in lookup_dict.items():
            if isinstance(value, list):
                for subvalue in value:
                    inverted_lookup_dict[subvalue] = key
            elif isinstance(value, int):
                inverted_lookup_dict[value] = key

        simple_to_complex = lookup_dict
        complex_to_simple = inverted_lookup_dict

        logging.info("Lookup tables from complex to simple and vice versa were created")
        return simple_to_complex, complex_to_simple



import re

import geopandas as gpd
import pandas as pd

from ra2ce.analysis.damages.shape_to_integrate_object.to_Integrate_shaper_protocol import (
    ToIntegrateShaperProtocol,
)


class OsdToIntegrateShaper(ToIntegrateShaperProtocol):
    gdf: gpd.GeoDataFrame

    def __init__(self, gdf):
        self.gdf = gdf

    @staticmethod
    def _extract_columns_by_pattern(
        pattern_text: str, gdf: gpd.GeoDataFrame
    ) -> set[str]:
        """
        Extract column names based on the provided regex pattern.

        Args:
            pattern_text (Pattern[str]): The compiled regex pattern to match the column names.
            df (pd.DataFrame): The DataFrame from which to extract the RP values.

        Returns:
            Set[str]: A set of RP values extracted from the column names.
        """
        pattern = re.compile(pattern_text)
        columns = {pattern.search(c).group(1) for c in gdf.columns if pattern.search(c)}
        return columns

    def get_return_periods(self) -> list:
        # Extract the RP values from the columns
        rp_values = OsdToIntegrateShaper._extract_columns_by_pattern(
            pattern_text=r"RP(\d+)", gdf=self.gdf
        )
        if not rp_values:
            raise ValueError("No damage column with RP found")

        return sorted([float(rp) for rp in rp_values])

    def shape_to_integrate_object(
        self, return_periods: list
    ) -> dict[str : gpd.GeoDataFrame]:
        _to_integrate_dict = {}
        # finds vulnerability curves shown with Ci, where 1<= i <=6
        vulnerability_curves = sorted(
            OsdToIntegrateShaper._extract_columns_by_pattern(
                pattern_text=r"(C\d+)", gdf=self.gdf
            )
        )
        for vulnerability_curve in vulnerability_curves:

            filtered_columns = sorted(
                OsdToIntegrateShaper._extract_columns_by_pattern(
                    pattern_text=rf"(.*{vulnerability_curve}.*representative.*)",
                    gdf=self.gdf,
                )
            )
            _to_integrate = self.gdf[filtered_columns]

            _to_integrate.columns = [
                float(c.split("_")[2].replace("RP", "")) for c in _to_integrate.columns
            ]
            _to_integrate_dict[f"{vulnerability_curve}"] = _to_integrate.sort_index(
                axis="columns", ascending=False
            )

        return _to_integrate_dict



from typing import Any

from ra2ce.analysis.analysis_config_data.enums.weighing_enum import WeighingEnum
from ra2ce.analysis.losses.weighing_analysis.weighing_analysis_protocol import (
    WeighingAnalysisProtocol,
)


class LengthWeighingAnalysis(WeighingAnalysisProtocol):
    edge_data: dict[str, Any]

    def __init__(self) -> None:
        self.edge_data = {}

    def _calculate_distance(self, time: float) -> float:
        _avgspeed = self.edge_data.get("avgspeed", None)  # km/h
        return round(time * _avgspeed * 1e3, 1)  # m

    def get_current_value(self) -> float:
        _dist = self.edge_data.get(WeighingEnum.LENGTH.config_value, None)  # h
        if _dist:
            return round(_dist, 1)
        _time = self.edge_data.get(WeighingEnum.TIME.config_value, 0)  # m
        _dist = self._calculate_distance(_time)
        self.edge_data[WeighingEnum.LENGTH.config_value] = _dist
        return _dist



"""
                    GNU GENERAL PUBLIC LICENSE
                      Version 3, 29 June 2007

    Risk Assessment and Adaptation for Critical Infrastructure (RA2CE).
    Copyright (C) 2023 Stichting Deltares

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from __future__ import annotations

import math
from dataclasses import dataclass, field
from pathlib import Path
from typing import Optional

from ra2ce.analysis.analysis_config_data.enums.analysis_damages_enum import (
    AnalysisDamagesEnum,
)
from ra2ce.analysis.analysis_config_data.enums.analysis_losses_enum import (
    AnalysisLossesEnum,
)
from ra2ce.analysis.analysis_config_data.enums.damage_curve_enum import DamageCurveEnum
from ra2ce.analysis.analysis_config_data.enums.event_type_enum import EventTypeEnum
from ra2ce.analysis.analysis_config_data.enums.loss_type_enum import LossTypeEnum
from ra2ce.analysis.analysis_config_data.enums.risk_calculation_mode_enum import (
    RiskCalculationModeEnum,
)
from ra2ce.analysis.analysis_config_data.enums.traffic_period_enum import (
    TrafficPeriodEnum,
)
from ra2ce.analysis.analysis_config_data.enums.trip_purpose_enum import TripPurposeEnum
from ra2ce.analysis.analysis_config_data.enums.weighing_enum import WeighingEnum
from ra2ce.common.configuration.config_data_protocol import ConfigDataProtocol
from ra2ce.network.network_config_data.enums.aggregate_wl_enum import AggregateWlEnum
from ra2ce.network.network_config_data.network_config_data import (
    NetworkSection,
    OriginsDestinationsSection,
)

LossesAnalysisNameList: list[str] = list(
    map(str, AnalysisLossesEnum.list_valid_options())
)
DamagesAnalysisNameList: list[str] = list(
    map(str, AnalysisDamagesEnum.list_valid_options())
)


@dataclass
class ProjectSection:
    """
    Reflects all possible settings that a project section might contain.
    """

    name: str = ""


@dataclass
class AnalysisSectionBase:
    """
    Reflects all common settings that damages and losses analysis sections might contain.
    """

    name: str = ""
    save_gpkg: bool = False
    save_csv: bool = False


@dataclass
class AnalysisSectionLosses(AnalysisSectionBase):
    """
    Reflects all possible settings that a losses analysis section might contain.
    """

    analysis: AnalysisLossesEnum = field(
        default_factory=lambda: AnalysisLossesEnum.INVALID
    )
    # general
    weighing: WeighingEnum = field(default_factory=lambda: WeighingEnum.NONE)
    loss_per_distance: str = ""
    loss_type: LossTypeEnum = field(default_factory=lambda: LossTypeEnum.NONE)
    disruption_per_category: str = ""
    # losses
    traffic_cols: list[str] = field(default_factory=list)
    production_loss_per_capita_per_hour: float = math.nan
    traffic_period: TrafficPeriodEnum = field(
        default_factory=lambda: TrafficPeriodEnum.DAY
    )
    hours_per_traffic_period: int = 0
    performance: str = "diff_time"  # "diff_time" or "diff_length" relates to the used criticality metric
    trip_purposes: list[TripPurposeEnum] = field(
        default_factory=lambda: [TripPurposeEnum.NONE]
    )
    resilience_curves_file: Optional[Path] = None
    traffic_intensities_file: Optional[Path] = None
    values_of_time_file: Optional[Path] = None
    # the redundancy analysis) and the intensities
    # accessibility analyses
    aggregate_wl: AggregateWlEnum = field(default_factory=lambda: AggregateWlEnum.NONE)
    threshold: float = 0.0
    threshold_destinations: float = math.nan
    uniform_duration: float = math.nan
    gdp_percapita: float = math.nan
    equity_weight: str = ""
    calculate_route_without_disruption: bool = False
    buffer_meters: float = math.nan
    threshold_locations: float = math.nan
    category_field_name: str = ""
    save_traffic: bool = False


@dataclass
class AnalysisSectionDamages(AnalysisSectionBase):
    """
    Reflects all possible settings that a damages analysis section might contain.
    """

    analysis: AnalysisDamagesEnum = field(
        default_factory=lambda: AnalysisDamagesEnum.INVALID
    )
    # adaptation/effectiveness measures
    return_period: float = math.nan
    repair_costs: float = math.nan
    evaluation_period: float = math.nan
    interest_rate: float = math.nan
    climate_factor: float = math.nan
    climate_period: float = math.nan
    # road damage
    representative_damage_percentage: float = 100
    event_type: EventTypeEnum = field(default_factory=lambda: EventTypeEnum.INVALID)
    damage_curve: DamageCurveEnum = field(
        default_factory=lambda: DamageCurveEnum.INVALID
    )
    risk_calculation_mode: RiskCalculationModeEnum = field(
        default_factory=lambda: RiskCalculationModeEnum.NONE
    )
    risk_calculation_year: int = 0
    create_table: bool = False
    file_name: Optional[Path] = None


@dataclass
class AnalysisConfigData(ConfigDataProtocol):
    """
    Reflects all config data from analysis.ini with defaults set.
    Additionally, some attributes from the network config are added for completeness (files, origins_destinations, network, hazard_names)
    """

    root_path: Optional[Path] = None
    input_path: Optional[Path] = None
    output_path: Optional[Path] = None
    static_path: Optional[Path] = None
    project: ProjectSection = field(default_factory=ProjectSection)
    analyses: list[AnalysisSectionBase] = field(default_factory=list)
    origins_destinations: Optional[OriginsDestinationsSection] = field(
        default_factory=OriginsDestinationsSection
    )
    network: NetworkSection = field(default_factory=NetworkSection)
    hazard_names: list[str] = field(default_factory=list)

    @property
    def damages_list(self) -> list[AnalysisSectionDamages]:
        """
        Get all damages analyses from config.

        Returns:
            list[AnalysisSectionDamages]: List of all damages analyses.
        """
        return list(
            filter(lambda x: isinstance(x, AnalysisSectionDamages), self.analyses)
        )

    @property
    def losses_list(self) -> list[AnalysisSectionLosses]:
        """
        Get all losses analyses from config.

        Returns:
            list[AnalysisSectionLosses]: List of all losses analyses.
        """
        return list(
            filter(lambda x: isinstance(x, AnalysisSectionLosses), self.analyses)
        )

    @staticmethod
    def get_data_output(ini_file: Path) -> Path:
        return ini_file.parent.joinpath("output")
